/*
Copyright (c) <2012>, Intel Corporation All Rights Reserved.

The source code, information and material ("Material") contained herein is owned by Intel Corporation or its suppliers or licensors, and title to such Material remains with Intel Corporation or its suppliers or licensors. The Material contains proprietary information of Intel or its suppliers and licensors. The Material is protected by worldwide copyright laws and treaty provisions. No part of the Material may be used, copied, reproduced, modified, published, uploaded, posted, transmitted, distributed or disclosed in any way without Intel's prior express written permission. No license under any patent, copyright or other intellectual property rights in the Material is granted to or conferred upon you, either expressly, by implication, inducement, estoppel or otherwise. Any license under such intellectual property rights must be express and approved by Intel in writing.

Copyright (c) Microsoft Corporation. All rights reserved

Unless otherwise agreed by Intel in writing, you may not remove or alter this notice or any other notice embedded in Materials by Intel or Intels suppliers or licensors in any way.
*/

#define STRICT
#define _WIN32_WINNT 0x501
#define _WIN32_IE 0x600

#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <string>
#include <vector>
#include <tchar.h>
using namespace std;
#pragma warning(disable : 4996)
#pragma warning(disable : 4267)

//
// Win8 Sensor Explorer
// by Dale Taylor, Intel Corp.
//	Available Sensors are displayed by Category
//	+Status and sensor information is displayed
//  +Sensor display is update on the fly
//

#include <sensorsapi.h>
#include <sensors.h>

#pragma comment(lib,"sensorsapi.lib")

// Globals
HWND MainWindow = 0;
HINSTANCE hAppInstance = 0;
HICON hIcon1;
const TCHAR* ttitle = _T("Win8 Sensor Explorer by Intel");

//debug variable for counting loops
int	dbloop = 0;
int dbloop2 = 0;

// in the COM interface this is the method for discovering and retreiving available sensors
ISensorManager* pISM = 0;

// function definitions that enable forward references to link
void UpdateSensorDatainList(int i, int j, int k, int lvi);
void AddSensorDatatoList(int i, int j, int k);
void UpdateData(ISensor* s,ISensorDataReport* pData);

const LPWSTR sens_categories[] = {L"Location",L"Environmental",L"Motion",L"Orientation",L"Mechanical",L"Electrical",L"Biometric",L"Light",L"Scanner"};
const CLSID sens_clsids[] = {SENSOR_CATEGORY_LOCATION,SENSOR_CATEGORY_ENVIRONMENTAL,SENSOR_CATEGORY_MOTION,SENSOR_CATEGORY_ORIENTATION,SENSOR_CATEGORY_MECHANICAL,SENSOR_CATEGORY_ELECTRICAL,SENSOR_CATEGORY_BIOMETRIC,SENSOR_CATEGORY_LIGHT,SENSOR_CATEGORY_SCANNER};
#define NUM_CATEGORIES 9	// number of sensor categories

// Sensor Connection Types as found in SensorConnectionType 0-2
const LPWSTR sens_ctype[] = {L"Integrated",L"Attached",L"External"};
#define NUM_CTYPES 3

//Sensor States as defined in the API
const LPWSTR sens_states[] = {L"Ready",L"Not Available",L"No Data",L"Initializing",L"Access Denied"L"Error"};
#define NUM_STATES 6		// max number of sensor states
#define MAX_SENSORTYPES 9	// max number of sensor types per category
#define MAX_DATAITEMS 9		// max number of sensor data fields per sensor

// This is the callback interface for receiving Sensor Manager Events
class MySensorManagerEvents : public ISensorManagerEvents
	{
	private:
		unsigned long ref;

	public:

		MySensorManagerEvents()
			{
			ref = 0;
			AddRef();
			}

		// IUnknown
		ULONG __stdcall AddRef()
			{
			return ++ref;
			}

		ULONG __stdcall Release()
			{
			return --ref;
			}
		HRESULT __stdcall QueryInterface(const IID& id,void**p)
			{
			if (id == IID_IUnknown || id == __uuidof(ISensorManagerEvents))
				{
				*p = this;
				AddRef();
				return S_OK;
				}
			return E_NOINTERFACE;
			}

		HRESULT __stdcall OnSensorEnter( ISensor *pSensor, SensorState state)
			{
			// Data changed, update
			SendMessage(MainWindow,WM_USER + 1,0,0);
			return S_OK;
			}

	};
MySensorManagerEvents SME;

// Define all of the COM structures to make things magically line up and work
// from the examples provided by Microsoft here:
// http://msdn.microsoft.com/en-us/library/windows/desktop/dd319014(v=vs.85).aspx
//
// This is the callback interface for receiving ISensorEvents
class CMySensorEvents : public ISensorEvents
	{
	private:
		unsigned long ref;

	public:
		CMySensorEvents()
			{
			ref = 0;
			AddRef();
			}

		// IUnknown
		ULONG __stdcall AddRef()
			{
			return ++ref;
			}

		ULONG __stdcall Release()
			{
			return --ref;
			}
		HRESULT __stdcall QueryInterface(const IID& id,void**p)
			{
			if (id == IID_IUnknown || id == __uuidof(ISensorEvents))
				{
				*p = this;
				AddRef();
				return S_OK;
				}
			return E_NOINTERFACE;
			}

		// Sensor Events  
		HRESULT __stdcall OnEvent(ISensor *pSensor, REFGUID eventID, IPortableDeviceValues *pEventData)
		{
			// handle custom events here
			return S_OK;
		}
		HRESULT __stdcall OnDataUpdated(ISensor *pSensor, ISensorDataReport *pNewData)
		{
			if (pSensor == NULL || pNewData == NULL)
				return E_INVALIDARG;
			UpdateData(pSensor, pNewData);
			return S_OK;
		}
		// sensor going away (unplugged) refresh the list
		HRESULT __stdcall OnLeave(REFSENSOR_ID sensorID)
		{
			HWND hL = GetDlgItem(MainWindow,901);
			ListView_DeleteAllItems(hL);			// clear the list
			PostMessage(MainWindow,WM_USER + 1,0,0);
			return S_OK;
		}
		// sensor state changed, refresh the list
		HRESULT __stdcall OnStateChanged(ISensor *pSensor, SensorState state)
		{
			HWND hL = GetDlgItem(MainWindow,901);
			ListView_DeleteAllItems(hL);			// clear the list
			PostMessage(MainWindow,WM_USER + 1,0,0);
			return S_OK;
		}

	};

// Structs used to keep local copies of the sensors for display
//
// This is the data for an individual sensor
struct SENSOR_DATA
	{
	PROPERTYKEY key;
	PROPVARIANT data;
	SYSTEMTIME	time;
	// the data below is for display purposes in the list
	int			lvidx;			// list view index number for this data (2nd list)
	};

// This is the data for a group of sensors
// in this program I group the sensors by category as I process through them all
// - several alternate methods of browsing the sensor data are also possible
struct SENSOR_GROUP
	{
	ISensor* pSen;
	CMySensorEvents* pCMSE;			// sensor event class w/handler
	CLSID	sclsid;					// category for this sensor
	int		count;					// number of sensor data items in the list
	SENSOR_DATA xData[MAX_DATAITEMS];
	};

// This is the base object containing all data for all sensors on this system
// One sensor List for each Category
struct SENSOR_LIST
	{
	int count;						// number of sensor items in the list
	SENSOR_GROUP xSensor[MAX_SENSORTYPES];
	};

// define an array of sensors, where we fill in the data by category
SENSOR_LIST		xSenCat[NUM_CATEGORIES];		// define an array of sensors 
int SCount = 0;								// we start with no categories of sensors present

// update the sensor data in our struct for a single sensor
// after updated the data, we then update the list for display
void UpdateSingleSensor(ISensor* pSen, ISensorDataReport* pData, int i, int j)
{
	HRESULT hr = 0;

	IPortableDeviceKeyCollection* pDKC = NULL;

	hr = pSen->GetSupportedDataFields(&pDKC);		// use this sensor to look at the data fields
	if (hr == S_OK)									// see if we got any data?
	{
		PROPERTYKEY pk;
		PROPVARIANT pv;
		SYSTEMTIME	ts;
		ULONG cnt = 0;

		xSenCat[i].xSensor[j].count = 0;				// default to no sensors
		hr = pDKC->GetCount(&cnt);						// get # of keys for this sensor type
	
		if (SUCCEEDED(hr))								// if we didn't get any data, skip
		{
			// loop through each of our data properties and get the values
			for(unsigned int k = 0 ; k < cnt ; k++)			// k is index into this sensors properties
			{
				pDKC->GetAt(k,&pk);							// for each field, get the keyname (used to look up the data)
				pData->GetSensorValue(pk,&pv);				// get the sensor value for this key
				pData->GetTimestamp(&ts);					// Get the time stamp from the event

				xSenCat[i].xSensor[j].xData[k].key = pk;	// save the key in the struct
				xSenCat[i].xSensor[j].xData[k].data = pv;	// save the data in the struct
				xSenCat[i].xSensor[j].xData[k].time = ts;	// save the timestamp in the struct

				// update the individual sensor in the list
				UpdateSensorDatainList(i, j, k, xSenCat[i].xSensor[j].xData[k].lvidx);
			}
			xSenCat[i].xSensor[j].count = cnt;				// set the # of keys for this sensor
		}
		pDKC->Release();
	}
}

// update the data fields for this sensor in the list view
//
void UpdateData(ISensor* pSen, ISensorDataReport* pData)
{
	HRESULT hr = 0;
	HWND hh = MainWindow;
	HWND hL = GetDlgItem(MainWindow,901);
	HWND hL2 = GetDlgItem(hh,902);

	if (pSen)						// do we have a sensor to update?
	{
		// loop for each sensor category in the struture
		for(int i = 0 ; i < NUM_CATEGORIES; i++)
		{
			// look at each sensor type in this category
			for (int j = 0; j < xSenCat[i].count; j++)
			{
				if (pSen == xSenCat[i].xSensor[j].pSen)		// is this the sensor we want?
				{
					UpdateSingleSensor(pSen, pData, i, j);	// -yes, update it then

					// tell the window manager that we need to update now
					NMHDR nh;
					nh.code = LVN_ITEMCHANGED;
//					nh.hwndFrom = hL;				// set the window to update
					nh.hwndFrom = hL2;				// set the window to update
					SendMessage(MainWindow,WM_NOTIFY,0,(LPARAM)&nh);
					return;
				}
			} // index through all sensors in each category [j]
		} // index through all xSenCat categories [i]
	}
}

// This needs to be called once and only once at startup
// inventory the sensors, and load the info into a struct for display
//	i is the loop counter for the number of categories
//	j is the loop counter for the number of sensors in the category
//  k is the loop counter for each property in a sensor
void LoadSensors()
{
	HWND hh = MainWindow;
	HWND hL = GetDlgItem(MainWindow,901);
	HWND hL2 = GetDlgItem(MainWindow,902);
	HRESULT hr = 0;
	wchar_t dtemp[128];			// allocate a character buffer to build strings in
	ULONG cnt = 0;

//	::MessageBox(NULL,_T("DT:in LoadSensors"), _T("DT:Debug Message"), MB_OK | MB_ICONERROR);

	ListView_DeleteAllItems(hL);			// clear the list

	// loop for each type of sensor Category
	for(int i = 0 ; i < NUM_CATEGORIES ; i++)
	{
		ISensorCollection* pSCol = 0;
		xSenCat[i].count = cnt = 0;				// start w/no sensors in this category

		hr = pISM->GetSensorsByCategory(sens_clsids[i],&pSCol);	// Get the sensor collection for this category
		if (!pSCol)				// if no sensors in this category, skip to next category
			goto Exit2;			// fill in the display info before leaving

//		hr = sm->RequestPermissions(0,pSenCol,TRUE);

		hr = pSCol->GetCount(&cnt);				//  Get # of sensors in this category to look at
		if (!SUCCEEDED(hr))						// if we didn't get any data, skip
			goto Exit1;							// -no, skip to next category
		if (!cnt)								// any sensors in this category?
		{
			goto Exit1;						// -no, skip to next category
		}
		xSenCat[i].count = cnt;				// -yes, save the number of sensors

		// load up all the data for each sensor in this category
		for(unsigned int j = 0 ; j < cnt ; j++)
		{
			ISensor* pSen = 0;						// define a pointer to an iSensor
			CMySensorEvents* pEventClass = NULL;	// pointer to sensor class
			ISensorEvents* pMyEvents = NULL;

			hr = pSCol->GetAt(j,&pSen);				// get the address in the pointer
			if (hr != S_OK)							// if this sensor isn't defined, skip it
			{
				xSenCat[i].xSensor[j].pSen = 0;		// mark the sensor as null
				continue;
			}

// event setup as outlined by Microsoft here:
// http://msdn.microsoft.com/en-us/library/windows/desktop/dd319014(v=vs.85).aspx
			// for EACH sensor, setup the event handler
			pEventClass = new CMySensorEvents();		// create an instance of the event class
			hr = pEventClass->QueryInterface(IID_PPV_ARGS(&pMyEvents));	// get pointer to the callback interface
			hr = pSen->SetEventSink(pMyEvents);			// set the event handler using ptr received

			xSenCat[i].xSensor[j].pCMSE = pEventClass;	// save the reference in the struct
			xSenCat[i].xSensor[j].pSen = pSen;			// save the sensor pointer in the struct

			GUID pguid[4];
			pguid[3] = SENSOR_EVENT_PROPERTY_CHANGED;
			pguid[2] = SENSOR_EVENT_ACCELEROMETER_SHAKE;
			pguid[1] = SENSOR_EVENT_DATA_UPDATED;
			pguid[0] = SENSOR_EVENT_STATE_CHANGED;
			hr = pSen->SetEventInterest(pguid,3);		// for EACH sensor, setup the events we are interested in

			ISensorDataReport* pSDR = NULL;

			hr = xSenCat[i].xSensor[j].pSen->GetData(&pSDR);	// we have a sensor, get the data
			if (SUCCEEDED(hr))							// if we didn't get any data, skip
			{
				UpdateSingleSensor(xSenCat[i].xSensor[j].pSen, pSDR, i, j);
				pSDR->Release();	// if we got sensor data memory, then release it
			}

			// Now lets build the list display so we can see how things are going
			int L = ListView_GetItemCount(hL);			// get # of items currently in the list
			_stprintf(dtemp,_T("%02u"),L + 1);			// convert item # to a string we'll display

			LV_ITEM lv = {LVIF_TEXT  | LVIF_GROUPID,L,0,0,0,dtemp,0,0,0};	// define a ListView item, set it up with defaults
			lv.iGroupId = i;							// put category id in the list
            ListView_InsertItem(hL,&lv);				// insert a line into the list (increases item count)
			ListView_SetItemText(hL,L,2,sens_categories[i]);		// Set the sensor category text for this line in the table

			ListView_GetItemText(hL,i,5,dtemp, 128);	// get any existing text
			_stprintf(dtemp,_T("%s+%u"),dtemp, xSenCat[i].xSensor[j].count);	// convert data to a string, adding additional info
			ListView_SetItemText(hL,i,5,dtemp);			// set the item count in the list view

			// Get the name of the sensor and add it to the ListView
			BSTR bSname = 0;		// temp string
			hr = pSen->GetFriendlyName(&bSname);
			if (bSname)
				{
				ListView_SetItemText(hL,L,1,bSname);
				SysFreeString(bSname);
				}

			// get the state for the current sensor and add it to the ListView
			SensorState st;
			hr = pSen->GetState(&st);
			if (st == SENSOR_STATE_ACCESS_DENIED)
				{
				hr = pISM->RequestPermissions(hh,pSCol,true);
				hr = pSen->GetState(&st);
				}
			switch(st)
				{
				case SENSOR_STATE_READY:
					{
					ListView_SetItemText(hL,L,3,L"Ready");
					break;
					}
				case SENSOR_STATE_NOT_AVAILABLE:
					{
					ListView_SetItemText(hL,L,3,L"Not Available");
					break;
					}
				case SENSOR_STATE_INITIALIZING:
					{
					ListView_SetItemText(hL,L,3,L"Initializing");
					break;
					}
				case SENSOR_STATE_NO_DATA:
					{
					ListView_SetItemText(hL,L,3,L"No Data");
					break;
					}
				case SENSOR_STATE_ERROR:
					{
					ListView_SetItemText(hL,L,3,L"Error");
					break;
					}
				case SENSOR_STATE_ACCESS_DENIED:
					{
					ListView_SetItemText(hL,L,3,L"Disabled");
					break;
					}
				}

			// loop through each of our data fields (properties) and get the values
			for(int k = 0 ; k < xSenCat[i].xSensor[j].count ; k++)
			{
				dbloop++;	//debug loop counter
				AddSensorDatatoList(i, j, k);	// add this sensor to the list
			}
		} // j is the index into the sensors 
		SCount++;							// we have a sensor, account for it
Exit1:	pSCol->Release();					// free the container memory
Exit2:	_stprintf(dtemp,_T("%2u"),xSenCat[i].count);	// convert sensor count to a string we can display
		ListView_SetItemText(hL,i,4,dtemp);				// set the item count in the list view
	}

	// format each of the ListView columns and set the display width
	for (int i = 0 ; i < 10 ; i++)
		SendMessage(hL, LVM_SETCOLUMNWIDTH, i, LVSCW_AUTOSIZE_USEHEADER);

	// resize the colums to fit the data in the display
	for(int i = 0 ; i < 3 ; i++)
		SendMessage(hL2, LVM_SETCOLUMNWIDTH, i, LVSCW_AUTOSIZE_USEHEADER);
}

//
// Creates a new list entry and then adds the sensor data to that position
//
void AddSensorDatatoList(int i, int j, int k)
{
	HWND hh = MainWindow;
	HWND hL = GetDlgItem(MainWindow,901);
	HWND hL2 = GetDlgItem(MainWindow,902);

	wchar_t	wtemp[128];					// temp buffer for building strings

	_stprintf(wtemp,_T("%02u"),i + 1);				// format Sensor number for display (# used in the first list)

	int lvi = ListView_GetItemCount(hL2);			// Get # of items currently in the 2nd list

	LV_ITEM lv = {LVIF_TEXT,lvi,0,0,0,wtemp,0,0,0};	// create an item for the list
	ListView_InsertItem(hL2,&lv);					// insert this into the 2nd list

	// save the item position in the list with the sensor info for updating the display
	xSenCat[i].xSensor[j].xData[k].lvidx = lvi;

	UpdateSensorDatainList(i, j, k, lvi);				// now update the list entry
}

//
// Get the sensor data item as pointed to by the parameters
// format the event and place it in the current list
//
void UpdateSensorDatainList(int i, int j, int k, int lvi)
{
	HWND hh = MainWindow;
	HWND hL = GetDlgItem(MainWindow,901);
	HWND hL2 = GetDlgItem(MainWindow,902);

	wchar_t	wtemp[128];					// temp buffer for building strings

	PROPERTYKEY& key = xSenCat[i].xSensor[j].xData[k].key;		// get the key
	PROPVARIANT& val = xSenCat[i].xSensor[j].xData[k].data;		// get the value
	SYSTEMTIME& stm = xSenCat[i].xSensor[j].xData[k].time;		// get the time

	// go through each of the sensor data types and see which this is
	// all categories GUID and PID info is found here:
	// http://msdn.microsoft.com/en-us/library/windows/desktop/dd318969(v=vs.85).aspx
	if (key == SENSOR_DATA_TYPE_LIGHT_CHROMACITY)
		{
		ListView_SetItemText(hL2,lvi,1,_T("Light Chromacity"));
		goto Exit1;
		}
	else
	if (key == SENSOR_DATA_TYPE_LIGHT_LEVEL_LUX)
		{
		ListView_SetItemText(hL2,lvi,1,_T("Lux Level"));
		goto Exit1;
		}
	else
	if (key == SENSOR_DATA_TYPE_LIGHT_TEMPERATURE_KELVIN)
		{
		ListView_SetItemText(hL2,lvi,1,_T("Light Temperature"));
		goto Exit1;
		}
	else
	// Motion PID's along with GUID info found here:
	// http://msdn.microsoft.com/en-us/library/windows/desktop/dd318985(v=vs.85).aspx
	if (key == SENSOR_DATA_TYPE_ACCELERATION_X_G)		// PID: 02 accelerationg_X_G
		{
		ListView_SetItemText(hL2,lvi,1,_T("Motion: Accel X-G"));
		goto Exit1;
		}
	else
	if (key == SENSOR_DATA_TYPE_ACCELERATION_Y_G)		// PID: 03 accelerationg_Y_G
		{
		ListView_SetItemText(hL2,lvi,1,_T("Motion: Accel Y-G"));
		goto Exit1;
		}
	else
	if (key == SENSOR_DATA_TYPE_ACCELERATION_Z_G)		// PID: 04 accelerationg_Z_G
		{
		ListView_SetItemText(hL2,lvi,1,_T("Motion: Accel Z-G"));
		goto Exit1;
		}
	else
	if (key == SENSOR_DATA_TYPE_ANGULAR_VELOCITY_X_DEGREES_PER_SECOND)		// PID: 10
		{
		ListView_SetItemText(hL2,lvi,1,_T("Motion: Angular DPS-X"));
		goto Exit1;
		}
	else
	if (key == SENSOR_DATA_TYPE_ANGULAR_VELOCITY_Y_DEGREES_PER_SECOND)		// PID: 11
		{
		ListView_SetItemText(hL2,lvi,1,_T("Motion: Angular DPS-Y"));
		goto Exit1;
		}
	else
	if (key == SENSOR_DATA_TYPE_ANGULAR_VELOCITY_Z_DEGREES_PER_SECOND)		// PID: 12
		{
		ListView_SetItemText(hL2,lvi,1,_T("Motion: Angular DPS-Z"));
		goto Exit1;
		}
	else
	// Location PID's along with GUID info found here:
	// http://msdn.microsoft.com/en-us/library/windows/desktop/dd318981(v=vs.85).aspx
	if (key == SENSOR_DATA_TYPE_LATITUDE_DEGREES)			// PID: 02	North is positive
		{
		ListView_SetItemText(hL2,lvi,1,_T("Location: Latitude Degrees"));
		_stprintf(wtemp,_T("LatDeg %f"), val.fltVal);
		goto Exit;
		}
	else
	if (key == SENSOR_DATA_TYPE_LONGITUDE_DEGREES)			// PID:03	East is positive
		{
		ListView_SetItemText(hL2,lvi,1,_T("Location: Longitude Degrees"));
		_stprintf(wtemp,_T("LonDeg %f"),val.fltVal);
		goto Exit;
		}
	else
	if (key == SENSOR_DATA_TYPE_ALTITUDE_ELLIPSOID_METERS)	// PID: 05	WGS 84 method
		{
		ListView_SetItemText(hL2,lvi,1,_T("Location: Altitude Elip"));
		_stprintf(wtemp,_T("Meters %f"), val.fltVal);
		goto Exit;
		}
	else
	if (key == SENSOR_DATA_TYPE_SPEED_KNOTS)				// PID: 06	
		{
		ListView_SetItemText(hL2,lvi,1,_T("Location: Speed in Knots"));
		_stprintf(wtemp,_T("Knots %f"), val.fltVal);
		goto Exit;
		}
	else
	if (key == SENSOR_DATA_TYPE_TRUE_HEADING_DEGREES)		// PID: 07
		{
		ListView_SetItemText(hL2,lvi,1,_T("Location: True Heading"));
		_stprintf(wtemp,_T("TrueN %f"), val.fltVal);
		goto Exit;
		}
	else
	if (key == SENSOR_DATA_TYPE_SATELLITES_IN_VIEW)			// PID:17
		{
		ListView_SetItemText(hL2,lvi,1,_T("Location: Sat in view"));
		_stprintf(wtemp,_T("%i"),val.intVal);
		goto Exit;
		}
	else
	if (key == SENSOR_DATA_TYPE_SATELLITES_IN_VIEW_PRNS)	// PID:18
		{
		ListView_SetItemText(hL2,lvi,1,_T("Location: Sats PRNS"));
		_stprintf(wtemp,_T("%i"),val.intVal);
		goto Exit;
		}
	else
	if (key == SENSOR_DATA_TYPE_SATELLITES_IN_VIEW_STN_RATIO)	// PID:21
		{
		ListView_SetItemText(hL2,lvi,1,_T("Location: Sat STN Ratio"));
		goto Exit1;
		}
	else
	if (key == SENSOR_DATA_TYPE_ERROR_RADIUS_METERS)		// PID:22
		{
		ListView_SetItemText(hL2,lvi,1,_T("Location: Error Radius Meters"));
		goto Exit1;
		}
	else
	if (key == SENSOR_DATA_TYPE_ALTITUDE_ELLIPSOID_ERROR_METERS)	// PID: 29
		{
		ListView_SetItemText(hL2,lvi,1,_T("Location: Alt Elip Error"));
		_stprintf(wtemp,_T("Meters %f"), val.fltVal);
		goto Exit;
		}
	else
// Orientation Sensors
	if (key == SENSOR_DATA_TYPE_TILT_X_DEGREES)						// PID: 02
		{
		ListView_SetItemText(hL2,lvi,1,_T("Orientation: Tilt-X"));
		_stprintf(wtemp,_T("Degrees %f"), val.fltVal);
		goto Exit;
		}
	else
	if (key == SENSOR_DATA_TYPE_TILT_Y_DEGREES)						// PID: 03
		{
		ListView_SetItemText(hL2,lvi,1,_T("Orientation: Tilt-Y"));
		_stprintf(wtemp,_T("Degrees %f"), val.fltVal);
		goto Exit;
		}
	else
	if (key == SENSOR_DATA_TYPE_TILT_Z_DEGREES)						// PID: 04
		{
		ListView_SetItemText(hL2,lvi,1,_T("Orientation: Tilt-Z"));
		_stprintf(wtemp,_T("Degrees %f"), val.fltVal);
		goto Exit;
		}
	else
//	if (key == SENSOR_DATA_TYPE_ANGULAR_VELOCITY_Y_DEGREES_PER_SECOND)	// PID: 11
	if (key == SENSOR_DATA_TYPE_MAGNETIC_HEADING_COMPENSATED_MAGNETIC_NORTH_DEGREES)	// PID: 11
		{
		ListView_SetItemText(hL2,lvi,1,_T("Orientation: Mag North"));
		_stprintf(wtemp,_T("Degrees %f"), val.fltVal);
		goto Exit;
		}
	else
	if (key == SENSOR_DATA_TYPE_ROTATION_MATRIX)					// PID: 16
		{
		ListView_SetItemText(hL2,lvi,1,_T("Orientation: device"));
		_stprintf(wtemp,_T("Device %f"), val.fltVal);
		goto Exit;
		}
	else
	if (key == SENSOR_DATA_TYPE_QUATERNION)							// PID: 17
		{
		ListView_SetItemText(hL2,lvi,1,_T("Orientation: quaternion"));
		_stprintf(wtemp,_T("Device %f"), val.fltVal);
		goto Exit;
		}
	else
	if (key == SENSOR_DATA_TYPE_SIMPLE_DEVICE_ORIENTATION)			// PID: 18
		{
		ListView_SetItemText(hL2,lvi,1,_T("Orientation: Quad 1-4"));
		_stprintf(wtemp,_T("Quadrant %f"), val.fltVal);
		goto Exit;
		}
	else
	if (key == SENSOR_DATA_TYPE_MAGNETIC_FIELD_STRENGTH_X_MILLIGAUSS)	// PID: 18
		{
		ListView_SetItemText(hL2,lvi,1,_T("Orientation: Mag Field X"));
		_stprintf(wtemp,_T("Milligauss %f"), val.fltVal);
		goto Exit;
		}
	else
	if (key == SENSOR_DATA_TYPE_MAGNETIC_FIELD_STRENGTH_Y_MILLIGAUSS)	// PID: 19
		{
		ListView_SetItemText(hL2,lvi,1,_T("Orientation: Mag Field Y"));
		_stprintf(wtemp,_T("Milligauss %f"), val.fltVal);
		goto Exit;
		}
	else
	if (key == SENSOR_DATA_TYPE_MAGNETIC_FIELD_STRENGTH_Z_MILLIGAUSS)	// PID: 20
		{
		ListView_SetItemText(hL2,lvi,1,_T("Orientation: Mag Field Z"));
		_stprintf(wtemp,_T("Milligauss %f"), val.fltVal);
		goto Exit;
		}
	else
	if (key == SENSOR_DATA_TYPE_TIMESTAMP)
		{
		ListView_SetItemText(hL2,lvi,1,_T("Timestamp"));
		_stprintf(wtemp,_T("Timestamp Y:%d, M:%d, D:%d, H:%d, M:%d, s:%d, ms:%d"),
			stm.wYear, stm.wMonth, stm.wDay, stm.wHour, stm.wMinute, stm.wSecond, stm.wMilliseconds);
		goto Exit;
		}
	else
	// if we didn't find any sensor specifics, list the category
	if (key.fmtid == SENSOR_DATA_TYPE_COMMON_GUID)
		{
		ListView_SetItemText(hL2,lvi,1,_T("Common GUID"));
		}
	else
	if (key.fmtid == SENSOR_DATA_TYPE_ORIENTATION_GUID)
		{
		ListView_SetItemText(hL2,lvi,1,_T("Orientation GUID"));
		}
	else
	if (key.fmtid == SENSOR_DATA_TYPE_LOCATION_GUID)
		{
		ListView_SetItemText(hL2,lvi,1,_T("Location GUID"));
		}
	else
	if (key.fmtid == SENSOR_DATA_TYPE_MOTION_GUID)
		{
		ListView_SetItemText(hL2,lvi,1,_T("Motion GUID"));
		}

	// Display the GUID and PID so I can look it up manually and update the formatting fields above
	LPOLESTR st;
	StringFromCLSID(key.fmtid,&st);						// convert format ID GUID to a string
	_stprintf(wtemp,_T("%s, PID: %i"),st, key.pid);		// format the string into temp
	CoTaskMemFree(st);
	goto Exit;

Exit1:
	_stprintf(wtemp,_T("%f"),val.fltVal);
Exit:
	ListView_SetItemText(hL2,lvi,2,wtemp);	// put the second string in the list

}

// main callback routine for the primary window
//
// hh = handle to window
// mm = message identifier
// ww = first message parameter
// ll = second message parameter
LRESULT CALLBACK MainWndProc(HWND hh,UINT mm,WPARAM ww,LPARAM ll)
{
	HWND hL = GetDlgItem(hh,901);
	HWND hL2 = GetDlgItem(hh,902);
	HRESULT hr = 0;
//	wchar_t	wtemp[128];					// temp buffer for building strings

	switch(mm)
	{
		case WM_USER + 1:
		{
//			::MessageBox(NULL,_T("Main_CB WM_USER"), _T("DT:Debug Message"), MB_OK | MB_ICONERROR);
//			UpdateData(0,0);	// setup Sensor data, something changed
			return 0;
		}

		case WM_NOTIFY:
		{
			NMHDR* n = (NMHDR*)ll;
			dbloop2++;

			if (n->hwndFrom == hL)		// make sure the notification is for our window
			{
				// do anything necessary to update your window here
			}
			return 0;
		}

		case WM_CREATE:
		{
			MainWindow = hh;
			hL = CreateWindowEx(0,_T("SysListView32"),_T(""),LVS_REPORT | WS_CHILD | WS_VISIBLE,0,0,100,100,hh,(HMENU)901,GetModuleHandle(0),0);
			hL2 = CreateWindowEx(0,_T("SysListView32"),_T(""),LVS_REPORT | WS_CHILD | WS_VISIBLE,0,0,100,100,hh,(HMENU)902,GetModuleHandle(0),0);

			// Build the list view we will fill with data here.
			const LPWSTR n[] = {L"#",L"Sensor Friendly Name",L"Category",L"State",L"#Sensors",L"#/Sensor"};
			for(int i = 0 ; i < 6 ; i++)
			{
				LV_COLUMN lv0 = {LVCF_SUBITEM | LVCF_TEXT | LVCF_WIDTH,0,70,n[i]};
				ListView_InsertColumn(hL,i,&lv0);
			}
			SendMessage(hL, LVM_FIRST + 54, LVS_EX_FULLROWSELECT | LVS_EX_GRIDLINES | LVS_EX_DOUBLEBUFFER,
				LVS_EX_FULLROWSELECT | LVS_EX_GRIDLINES | LVS_EX_DOUBLEBUFFER);

			const LPWSTR n2[] = {L"#",L"Sensor Description",L"Sensor Data Value"};
			for(int i = 0 ; i < 3 ; i++)
			{
				LV_COLUMN lv0 = {LVCF_SUBITEM | LVCF_TEXT | LVCF_WIDTH,0,70,n2[i]};
				ListView_InsertColumn(hL2,i,&lv0);
			}
			SendMessage(hL2, LVM_FIRST + 54, LVS_EX_FULLROWSELECT | LVS_EX_GRIDLINES | LVS_EX_DOUBLEBUFFER,
				LVS_EX_FULLROWSELECT | LVS_EX_GRIDLINES | LVS_EX_DOUBLEBUFFER);

			// Groups
			ListView_EnableGroupView(hL,true);
			for(int i = 0 ; i < NUM_CATEGORIES ; i++)
			{
				LVGROUP lg = {0};
				lg.cbSize = sizeof(LVGROUP);
				lg.mask = LVGF_HEADER | LVGF_GROUPID;
				lg.pszHeader = sens_categories[i];
				lg.iGroupId = i;
				ListView_InsertGroup(hL, -1, &lg);
			}

			CoCreateInstance(CLSID_SensorManager,0,CLSCTX_ALL,__uuidof(ISensorManager),(void**)&pISM);
			if (!pISM)
			{
				MessageBox(NULL,L"Sensor Manager not created. Check windows version?",ttitle,MB_OK);
				return 0;
			}
			hr = pISM->SetEventSink(&SME);		// hook the Sensor Manager Events to get the calls

			// we've created the window, now lets load up the sensor data in our structure
			LoadSensors();		// setup table with sensor names, set event handlers too!

			ShowWindow(hL,SW_HIDE);
			ShowWindow(hL2,SW_HIDE);

			// debug message
//			_stprintf(wtemp,_T("SCount = %i, DBLoop = %i, DBLoop2 = %i"), SCount, dbloop, dbloop2);			// format item number for display
//			MessageBox(NULL,wtemp,ttitle,MB_OK);

			SendMessage(hL, LVM_FIRST + 54, LVS_EX_FULLROWSELECT | LVS_EX_DOUBLEBUFFER | LVS_EX_GRIDLINES,
				  LVS_EX_FULLROWSELECT | LVS_EX_DOUBLEBUFFER | LVS_EX_GRIDLINES);
			ShowWindow(hL,SW_SHOW);
			SendMessage(hL2, LVM_FIRST + 54, LVS_EX_FULLROWSELECT | LVS_EX_DOUBLEBUFFER | LVS_EX_GRIDLINES,
				  LVS_EX_FULLROWSELECT | LVS_EX_DOUBLEBUFFER | LVS_EX_GRIDLINES);
			ShowWindow(hL2,SW_SHOW);

			SendMessage(hh,WM_SIZE,0,0);
			break;
		}

		case WM_COMMAND:
		{
			int LW = LOWORD(ww);

			return 0;
		}

		case WM_SIZE:
		{
			RECT rc;
			GetClientRect(hh,&rc);
			SetWindowPos(hL,0,0,0,rc.right/2,rc.bottom,SWP_SHOWWINDOW);
			SetWindowPos(hL2,0,rc.right/2,0,rc.right/2,rc.bottom,SWP_SHOWWINDOW);
			return 0;
		}

		case WM_CLOSE:
		{
			ListView_DeleteAllItems(hL);			// clear the list
			DestroyWindow(hh);
			pISM->Release();						// free all the sensor memory (pCMSE etc)
			pISM = 0;
			return 0;
		}

		case WM_DESTROY:
		{
			PostQuitMessage(0);
			return 0;
		}
	}
	return DefWindowProc(hh,mm,ww,ll);
}


int __stdcall WinMain(HINSTANCE h,HINSTANCE,LPSTR tt,int)
{
	CoInitializeEx(0,COINIT_APARTMENTTHREADED);
	INITCOMMONCONTROLSEX icex = {0};
	icex.dwICC = ICC_LISTVIEW_CLASSES | ICC_DATE_CLASSES | ICC_WIN95_CLASSES;
	icex.dwSize = sizeof(icex);
	InitCommonControlsEx(&icex);
	hIcon1 = LoadIcon(h,_T("Light_Bulb"));

	hAppInstance = h;

	WNDCLASSEX wClass = {0};
	wClass.cbSize = sizeof(wClass);
	
	wClass.style = CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW | CS_PARENTDC;
	wClass.lpfnWndProc = (WNDPROC)MainWndProc;
	wClass.hInstance = h;
	wClass.hIcon = hIcon1;
	wClass.hCursor = LoadCursor(0, IDC_ARROW);
	wClass.hbrBackground = (HBRUSH) GetStockObject(WHITE_BRUSH);
	wClass.lpszClassName = _T("CLASS");
	wClass.hIconSm = hIcon1;
	RegisterClassEx(&wClass);

	MainWindow = CreateWindowEx(0,
		_T("CLASS"),
		ttitle,
		WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS |
		WS_CLIPCHILDREN, CW_USEDEFAULT, CW_USEDEFAULT,
		CW_USEDEFAULT, CW_USEDEFAULT, 0,LoadMenu(h,L"MENU_1"), h, 0);

	ShowWindow(MainWindow,SW_SHOW);


	MSG msg;
	while(GetMessage(&msg,0,0,0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}


	return 0;
}

